  ' program demo_geodetic.bas      July 30, 2017
  
  ' this program demonstrates the proper method for
  ' interacting with several MMBASIC subroutines
  ' that perform geocentric to/from geodetic
  ' coordinate conversions  

  ' MMX, RasPi and DOS version
  
  ''''''''''''''''''''''''''''
  
  option default float
  
  option base 1
  
  dim rvector1(3), rvector2(3)
  
  const pidiv2 = 0.5 * pi
  
  ' conversion factor - degrees to radians
  
  const dtr = pi / 180.0
  
  ' conversion factor - radians to degrees
  
  const rtd = 180.0 / pi
  
  ' equatorial radius of the Earth (kilometers)
  
  req = 6378.14
  
  ' flattening factor of the Earth (non-dimensional)
  
  flat = 1.0 / 298.257
  
  print " "
  print "geodetic and geocentric coordinate conversions"
  print "----------------------------------------------"
  print " "
  
  do
    
    print "please input the geocentric radius (kilometers)"
    
    input rmag
    
    if (rmag > req) then exit do
    
  loop
  
  do
    
    print " "
    
    print "please input the geocentric declination (degrees)"
    
    input xdata
    
    if (xdata >= -90.0 and xdata <= 90.0) then exit do
    
  loop
  
  decl = dtr * xdata
  
  do
    
    print " "
    
    print "please input the geocentric east longitude (degrees)"
    
    input xdata
    
    if (xdata >= 0.0 and xdata <= 360.0) then exit do
    
  loop
  
  xlong = dtr * xdata
  
  rmsaved = rmag
  
  decsaved = decl
  
  geodet1(flat, req, rmag, decl, alt, xlat)
  
  print " "
  print " "
  print "sub geodet1.bas"
  print "(geocentric-to-geodetic - series solution)"
  print "------------------------------------------"
  print " "
  print "geocentric radius       =  ", str$(rmag, 0, 4) + " kilometers"
  print " "
  print "geocentric declination  =  ", str$(rtd * decl, 0, 4) + " degrees"
  print " "
  print "geodetic altitude       =  ", str$(alt, 0, 4) + " kilometers"
  print " "
  print "geodetic latitude       =  ", str$(rtd * xlat, 0, 4) + " degrees"
  
  geodet2(flat, req, decl, rmag, xlat, alt)
  
  print " "
  print " "
  print "sub geodet2.bas"
  print "(geocentric-to-geodetic - Borkowski solution)"
  print "---------------------------------------------"
  print " "
  print "geocentric radius       =  ", str$(rmag, 0, 4) + " kilometers"
  print " "
  print "geocentric declination  =  ", str$(rtd * decl, 0, 4) + " degrees"
  print " "
  print "geodetic altitude       =  ", str$(alt, 0, 4) + " kilometers"
  print " "
  print "geodetic latitude       =  ", str$(rtd * xlat, 0, 4) + " degrees"
  
  geodet3 (flat, req, xlat, xlong, alt, rvector1())
  
  print " "
  print " "
  print "sub geodet3.bas"
  print "(geodetic coordinates to ecf position vector)"
  print "---------------------------------------------"
  print " "
  print "ecf x-position          =  ", str$(rvector1(1), 0, 4) + " kilometers"
  print " "
  print "ecf y-position          =  ", str$(rvector1(2), 0, 4) + " kilometers"
  print " "
  print "ecf z-position          =  ", str$(rvector1(3), 0, 4) + " kilometers"
  print " "
  print "geocentric radius       =  ", str$(rmsaved, 0, 4) + " kilometers"
  print " "
  print "geocentric declination  =  ", str$(rtd * decsaved, 0, 4) + " degrees"
  print " "
  print "geodetic altitude       =  ", str$(alt, 0, 4) + " kilometers"
  print " "
  print "geodetic latitude       =  ", str$(rtd * xlat, 0, 4) + " degrees"
  print " "
  print "east longitude          =  ", str$(rtd * xlong, 0, 4) + " degrees"
  
  geodet4(flat, req, xlat, alt, rmag, decl)
  
  print " "
  print " "
  print "sub geodet4.bas"
  print "(geodetic-to-geocentric - series solution)"
  print "------------------------------------------"
  print " "
  print "geocentric radius       =  ", str$(rmag, 0, 4) + " kilometers"
  print " "
  print "geocentric declination  =  ", str$(rtd * decl, 0, 4) + " degrees"
  print " "
  print "geodetic altitude       =  ", str$(alt, 0, 4) + " kilometers"
  print " "
  print "geodetic latitude       =  ", str$(rtd * xlat, 0, 4) + " degrees"
  
  ' compute geocentric position vector (kilometers)
  
  rvector2(1) = rmsaved * cos(decsaved) * cos(xlong)
  
  rvector2(2) = rmsaved * cos(decsaved) * sin(xlong)
  
  rvector2(3) = rmsaved * sin(decsaved)
  
  ecf2geodet(flat, req, rvector2(), xlat, xlong, alt)
  
  print " "
  print " "
  print "sub ecf2geodet.bas"
  print "(ecf position vector-to-geodetic coordinates)"
  print "---------------------------------------------"
  print " "
  print "ecf x-position          =  ", str$(rvector2(1), 0, 4) + " kilometers"
  print " "
  print "ecf y-position          =  ", str$(rvector2(2), 0, 4) + " kilometers"
  print " "
  print "ecf z-position          =  ", str$(rvector2(3), 0, 4) + " kilometers"
  print " "
  print "geocentric radius       =  ", str$(rmsaved, 0, 4) + " kilometers"
  print " "
  print "geocentric declination  =  ", str$(rtd * decsaved, 0, 4) + " degrees"
  print " "
  print "geodetic altitude       =  ", str$(alt, 0, 4) + " kilometers"
  print " "
  print "geodetic latitude       =  ", str$(rtd * xlat, 0, 4) + " degrees"
  print " "
  print "east longitude          =  ", str$(rtd * xlong, 0, 4) + " degrees"
  print " "
  
end
  
  '''''''''''''''''''''''''''''''''''''''''''
  '''''''''''''''''''''''''''''''''''''''''''
  
sub geodet1 (flat, req, rmag, dec, alt, xlat)
  
  ' geodetic latitude and altitude sub
  
  ' series solution
  
  ' input
  
  '  flat = flattening factor (non-dimensional)
  '  req  = equatorial radius of central body
  '  rmag = geocentric distance (kilometers)
  '  dec  = geocentric declination (radians)
  
  ' output
  
  '  alt  = geodetic altitude (kilometers)
  '  xlat = geodetic latitude (radians)
  
  '''''''''''''''''''''''''''''''''''''
  
  local xn, o, a, p, q, r, s
  
  xn = req / rmag
  
  o = flat * flat
  
  a = 2.0 * dec
  p = sin(a)
  q = cos(a)
  
  a = 4.0 * dec
  r = sin(a)
  s = cos(a)
  
  ' geodetic latitude (radians)
  
  xlat = dec + flat * xn * p + o * xn * r * (xn - 0.25)
  
  ' geodetic altitude (km)
  
  alt = rmag + req * (flat * 0.5 * (1.0 - q) + o * (0.25 * xn - 0.0625) * (1.0 - s) - 1.0)
  
end sub
  
  '''''''''''''''''''''''''''''''''''''''''''
  '''''''''''''''''''''''''''''''''''''''''''
  
sub geodet2(flat, req, decl, rmag, xlat, alt)
  
  ' geocentric to geodetic coordinates
  
  ' exact solution (Borkowski, 1989)
  
  ' input
  
  '  decl = geocentric declination (radians)
  '  rmag = geocentric distance (kilometers)
  
  ' output
  
  '  xlat = geodetic latitude (radians)
  '  alt  = geodetic altitude (kilometers)
  
  ''''''''''''''''''''''''''''''''''''''''
  
  local fr, rx, rz, b, d, e, f, g, p, q, s, t, v
  
  fr = 1.0 / flat
  
  ' determine x and z components of the geocentric distance
  
  rx = rmag * cos(decl)
  
  rz = rmag * sin(decl)
  
  ' compute geodetic latitude and altitude
  
  if (rz = 0.0) then
    
    ' special case - equatorial
    
    xlat = 0.0
    
    alt = rmag - req
    
  elseif (abs(decl) = 0.5 * pi) then
    
    ' special case - polar
    
    xlat = decl
    
    alt = rmag - (req - req / fr)
    
  else
    
    ' general case
    
    b = sgn(rz) * (req - req / fr)
    
    e = ((rz + b) * b / req - req) / rx
    
    f = ((rz - b) * b / req + req) / rx
    
    p = (e * f + 1.0) * 4.0 / 3.0
    
    q = (e * e - f * f) * 2.0
    
    d = p * p * p + q * q
    
    if (d >= 0.0) then
      
      s = sqr(d) + q
      
      s = sgn(s) * (exp(log(abs(s)) / 3.0))
      
      v = p/s - s
      
      v = -(q + q + v * v * v)/(3.0 * p)
      
    else
      
      v = 2.0 * sqr(-p) * cos(acos(q / p / sqr(-p)) / 3.0)
      
    end if
    
    g = 0.5 * (e + sqr(e * e + v))
    
    t = sqr(g * g + (f - v * g) / (g + g - e)) - g
    
    xlat = atn((1.0 - t * t) * req / (2.0 * b * t))
    
    alt = (rx - req * t) * cos(xlat) + (rz - b) * sin(xlat)
    
  end if
  
end sub
  
  ''''''''''''''''''''''''''''''''''''''''''''
  ''''''''''''''''''''''''''''''''''''''''''''
  
sub geodet3 (flat, req, xlat, xlong, alt, r())
  
  ' convert geodetic coordinates to ecf position vector
  
  ' input
  
  '  xlat  = geodetic latitude (radians)
  '         (+north, -south -pi/2 <= xlat <= +pi/2)
  '  xlong = geodetic longitude (radians)
  '  alt   = geodetic altitude (kilometers)
  
  ' output
  
  '  r() = ecf position vector (kilometers)
  
  '''''''''''''''''''''''''''''''''''''''''
  
  local esqr, xn
  
  ' eccentricity squared
  
  esqr = 2.0 * flat - flat * flat
  
  xn = req / sqr(1.0 - esqr * sin(xlat) * sin(xlat))
  
  ' ecf position vector components
  
  r(1) = (xn + alt) * cos(xlat) * cos(xlong)
  
  r(2) = (xn + alt) * cos(xlat) * sin(xlong)
  
  r(3) = (xn * (1.0 - esqr) + alt) * sin(xlat)
  
end sub
  
  '''''''''''''''''''''''''''''''''''''''''''
  '''''''''''''''''''''''''''''''''''''''''''
  
sub geodet4 (flat, req, xlat, alt, rmag, dec)
  
  ' geodetic to geocentric coordinates
  
  ' input
  
  '  xlat = geodetic latitude (radians)
  '         (+north, -south -pi/2 <= lat <= +pi/2)
  '  alt  = geodetic altitude (kilometers)
  
  ' output
  
  '  rmag = geocentric position magnitude (kilometers)
  '  dec  = geocentric declination (radians)
  '         (+north, -south -pi/2 <= dec <= +pi/2)
  
  ''''''''''''''''''''''''''''''''''''''''''''''''
  
  local hhat, hp1, s2lat, c2lat, s4lat, c4lat, rho
  
  ' "normalize" altitude
  
  hhat = alt / req
  
  hp1 = hhat + 1.0
  
  ' calculate trig terms
  
  s2lat = sin(2.0 * xlat)
  
  c2lat = cos(2.0 * xlat)
  
  s4lat = sin(4.0 * xlat)
  
  c4lat = cos(4.0 * xlat)
  
  ' geocentric distance (same units as req)
  
  rho = hp1 + (0.5 * (c2lat - 1.0)) * flat + ((1.0 / (4.0 * hp1) + (1.0 / 16.0)) * (1.0 - c4lat)) * flat^2
  
  rmag = req * rho
  
  ' geocentric declination (radians)
  
  dec = xlat + (-s2lat / hp1) * flat + (-s2lat / (2.0 * hp1^2) + (1.0 / (4.0 * hp1^2) + 1.0 / (4.0 * hp1)) * s4lat) * flat^2
  
end sub
  
  '''''''''''''''''''''''''''''''''''''''''''''''''''
  '''''''''''''''''''''''''''''''''''''''''''''''''''
  
sub ecf2geodet (flat, req, rsite(), xlat, xlong, alt)
  
  ' convert geocentric position vector to geodetic coordinates
  
  ' input
  
  '  flat  = flattening factor (non-dimensional)
  '  req   = equatorial radius of central body
  '  rsite = ecf position vector (kilometers)
  
  ' output
  
  '  xlat  = geodetic latitude (radians)
  '          (+north, -south -pi/2 <= xlat <= -pi/2)
  '  xlong = east longitude (radians)
  '          (0 <= xlong <= pi/2)
  '  alt   = geodetic altitude (meters)
  '          (+ above sea level, - below sea level)
  
  '''''''''''''''''''''''''''''''''''''''''''''''''
  
  local rmag, esqr, p, r, xmu, tmp1, tmp2
  
  rmag = sqr(rsite(1)^2 + rsite(2)^2 + rsite(3)^2)
  
  ' east longitude (radians)
  
  xlong = atan3(rsite(2), rsite(1))
  
  esqr = 2.0 * flat - flat * flat
  
  p = sqr(rsite(1)^2 + rsite(2)^2)
  
  r = sqr(p^2 + rsite(3)^2)
  
  xmu = atn((rsite(3) / p) * (1.0 - flat + esqr * req / r))
  
  tmp1 = rsite(3) * (1.0 - flat) + esqr * req * sin(xmu)^3
  
  tmp2 = (1.0 - flat) * (p - esqr * req * cos(xmu)^3)
  
  ' geodetic latitude (radians)
  
  xlat = atn(tmp1 / tmp2)
  
  ' geodetic altitude (kilometers)
  
  alt = p * cos(xlat) + rsite(3) * sin(xlat) - req * sqr(1.0 - esqr * sin(xlat)^2)
  
end sub
  
  '''''''''''''''''''''''''''
  '''''''''''''''''''''''''''
  
function atan3(a, b) as float
  
  ' four quadrant inverse tangent function
  
  ' input
  
  '  a = sine of angle
  '  b = cosine of angle
  
  ' output
  
  '  atan3 = angle (0 =< atan3 <= 2 * pi; radians)
  
  ''''''''''''''''''''''''''''''''''''''''''''''''
  
  local c
  
  if (abs(a) < 1.0e-10) then
    
    atan3 = (1.0 - sgn(b)) * pidiv2
    
    exit function
    
  else
    
    c = (2.0 - sgn(a)) * pidiv2
    
  endif
  
  if (abs(b) < 1.0e-10) then
    
    atan3 = c
    
    exit function
    
  else
    
    atan3 = c + sgn(a) * sgn(b) * (abs(atn(a / b)) - pidiv2)
    
  endif
  
end function

